@@ -23,7 +23,6 @@ import oslo_log.log as logging
 
 import neutronclient.common.exceptions as q_exceptions
 
-from tricircle.common import az_ag
 import tricircle.common.client as t_client
 from tricircle.common import constants
 import tricircle.common.context as t_context
@@ -32,10 +31,14 @@ from tricircle.common.i18n import _
 from tricircle.common.i18n import _LE
 import tricircle.common.lock_handle as t_lock
 from tricircle.common.quota import QUOTAS
+from tricircle.common.scheduler import filter_scheduler
 from tricircle.common import utils
+from tricircle.common import xrpcapi
 import tricircle.db.api as db_api
 from tricircle.db import core
 from tricircle.db import models
+from tricircle.network import helper
+
 
 LOG = logging.getLogger(__name__)
 
@@ -47,68 +50,91 @@ class ServerController(rest.RestController):
 
     def __init__(self, project_id):
         self.project_id = project_id
-        self.clients = {'top': t_client.Client()}
+        self.clients = {constants.TOP: t_client.Client()}
+        self.helper = helper.NetworkHelper()
+        self.xjob_handler = xrpcapi.XJobAPI()
+        self.filter_scheduler = filter_scheduler.FilterScheduler()
 
-    def _get_client(self, pod_name='top'):
+    def _get_client(self, pod_name=constants.TOP):
         if pod_name not in self.clients:
             self.clients[pod_name] = t_client.Client(pod_name)
         return self.clients[pod_name]
 
-    def _get_all(self, context):
+    def _get_all(self, context, params):
+        filters = [{'key': key,
+                    'comparator': 'eq',
+                    'value': value} for key, value in params.iteritems()]
         ret = []
         pods = db_api.list_pods(context)
         for pod in pods:
             if not pod['az_name']:
                 continue
             client = self._get_client(pod['pod_name'])
-            servers = client.list_servers(context)
+            servers = client.list_servers(context, filters=filters)
             self._remove_fip_info(servers)
             ret.extend(servers)
         return ret
 
+    @staticmethod
+    def _construct_brief_server_entry(server):
+        return {'id': server['id'],
+                'name': server.get('name'),
+                'links': server.get('links')}
+
+    @staticmethod
+    def _transform_network_name(server):
+        if 'addresses' not in server:
+            return
+        keys = [key for key in server['addresses'].iterkeys()]
+        for key in keys:
+            value = server['addresses'].pop(key)
+            network_name = key.split('#')[1]
+            server['addresses'][network_name] = value
+        return server
+
     @expose(generic=True, template='json')
-    def get_one(self, _id):
+    def get_one(self, _id, **kwargs):
         context = t_context.extract_context_from_environ()
 
         if _id == 'detail':
-            return {'servers': self._get_all(context)}
+            return {'servers': [self._transform_network_name(
+                server) for server in self._get_all(context, kwargs)]}
 
         mappings = db_api.get_bottom_mappings_by_top_id(
             context, _id, constants.RT_SERVER)
         if not mappings:
-            pecan.abort(404, 'Server not found')
-            return
+            return utils.format_nova_error(
+                404, _('Instance %s could not be found.') % _id)
         pod, bottom_id = mappings[0]
         client = self._get_client(pod['pod_name'])
         server = client.get_servers(context, bottom_id)
         if not server:
-            pecan.abort(404, 'Server not found')
-            return
+            return utils.format_nova_error(
+                404, _('Instance %s could not be found.') % _id)
         else:
+            self._transform_network_name(server)
             return {'server': server}
 
     @expose(generic=True, template='json')
-    def get_all(self):
+    def get_all(self, **kwargs):
         context = t_context.extract_context_from_environ()
-        return {'servers': self._get_all(context)}
+        return {'servers': [self._construct_brief_server_entry(
+            server) for server in self._get_all(context, kwargs)]}
 
     @expose(generic=True, template='json')
     def post(self, **kw):
         context = t_context.extract_context_from_environ()
 
         if 'server' not in kw:
-            pecan.abort(400, 'Request body not found')
-            return
+            return utils.format_nova_error(
+                400, _('server is not set'))
 
-        if 'availability_zone' not in kw['server']:
-            pecan.abort(400, 'Availability zone not set')
-            return
-
-        pod, b_az = az_ag.get_pod_by_az_tenant(
-            context, kw['server']['availability_zone'], self.project_id)
+        az = kw['server'].get('availability_zone', '')
+        pod, b_az = self.filter_scheduler.select_destination(
+            context, az, self.project_id, pod_group='')
         if not pod:
-            pecan.abort(400, 'No pod bound to availability zone')
-            return
+            return utils.format_nova_error(
+                500, _('Pod not configured or scheduling failure'))
 
         t_server_dict = kw['server']
         self._process_metadata_quota(context, t_server_dict)
@@ -129,47 +155,50 @@ class ServerController(rest.RestController):
             security_groups = []
             for sg in kw['server']['security_groups']:
                 if 'name' not in sg:
-                    pecan.abort(404, 'Security group name not specified')
-                    return
+                    return utils.format_nova_error(
+                        400, _('Invalid input for field/attribute'))
                 if sg['name'] not in top_sg_map:
-                    pecan.abort(404,
-                                'Security group %s not found' % sg['name'])
-                    return
+                    return utils.format_nova_error(
+                        400, _('Unable to find security_group with name or id '
+                               '%s') % sg['name'])
                 security_groups.append(sg['name'])
         t_sg_ids, b_sg_ids, is_news = self._handle_security_group(
             context, pod, top_sg_map, security_groups)
 
+        server_body['networks'] = []
         if 'networks' in kw['server']:
-            server_body['networks'] = []
             for net_info in kw['server']['networks']:
                 if 'uuid' in net_info:
                     network = top_client.get_networks(context,
                                                       net_info['uuid'])
                     if not network:
-                        pecan.abort(400, 'Network not found')
-                        return
+                        return utils.format_nova_error(
+                            400, _('Network %s could not be '
+                                   'found') % net_info['uuid'])
 
-                    if not self._check_network_server_the_same_az(
-                            network, kw['server']['availability_zone']):
-                        pecan.abort(400, 'Network and server not in the same '
-                                         'availability zone')
-                        return
+                    if not self._check_network_server_az_match(
+                            context, network,
+                            kw['server']['availability_zone']):
+                        return utils.format_nova_error(
+                            400, _('Network and server not in the same '
+                                   'availability zone'))
 
                     subnets = top_client.list_subnets(
                         context, [{'key': 'network_id',
                                    'comparator': 'eq',
                                    'value': network['id']}])
                     if not subnets:
-                        pecan.abort(400, 'Network not contain subnets')
-                        return
+                        return utils.format_nova_error(
+                            400, _('Network does not contain any subnets'))
                     t_port_id, b_port_id = self._handle_network(
                         context, pod, network, subnets,
                         top_sg_ids=t_sg_ids, bottom_sg_ids=b_sg_ids)
                 elif 'port' in net_info:
                     port = top_client.get_ports(context, net_info['port'])
                     if not port:
-                        pecan.abort(400, 'Port not found')
-                        return
+                        return utils.format_nova_error(
+                            400, _('Port %s could not be '
+                                   'found') % net_info['port'])
                     t_port_id, b_port_id = self._handle_port(
                         context, pod, port)
                 server_body['networks'].append({'port': b_port_id})
@@ -213,229 +242,110 @@ class ServerController(rest.RestController):
                                   'pod_id': pod['pod_id'],
                                   'project_id': self.project_id,
                                   'resource_type': constants.RT_SERVER})
+        pecan.response.status = 202
         return {'server': server}
 
+    @expose(generic=True, template='json')
+    def delete(self, _id):
+        context = t_context.extract_context_from_environ()
+
+        mappings = db_api.get_bottom_mappings_by_top_id(context, _id,
+                                                        constants.RT_SERVER)
+        if not mappings:
+            pecan.response.status = 404
+            return {'Error': {'message': _('Server not found'), 'code': 404}}
+
+        pod, bottom_id = mappings[0]
+        client = self._get_client(pod['pod_name'])
+        top_client = self._get_client()
+        try:
+            server_ports = top_client.list_ports(
+                context, filters=[{'key': 'device_id', 'comparator': 'eq',
+                                   'value': _id}])
+            ret = client.delete_servers(context, bottom_id)
+            # none return value indicates server not found
+            if ret is None:
+                self._remove_stale_mapping(context, _id)
+                pecan.response.status = 404
+                return {'Error': {'message': _('Server not found'),
+                                  'code': 404}}
+            for server_port in server_ports:
+                self.xjob_handler.delete_server_port(context,
+                                                     server_port['id'])
+        except Exception as e:
+            code = 500
+            message = _('Delete server %(server_id)s fails') % {
+                'server_id': _id}
+            if hasattr(e, 'code'):
+                code = e.code
+            ex_message = str(e)
+            if ex_message:
+                message = ex_message
+            LOG.error(message)
+
+            pecan.response.status = code
+            return {'Error': {'message': message, 'code': code}}
+
+        # NOTE(zhiyuan) Security group rules for default security group are
+        # also kept until subnet is deleted.
+        pecan.response.status = 204
+        return pecan.response
+
     def _get_or_create_route(self, context, pod, _id, _type):
-        def list_resources(t_ctx, q_ctx, pod_, _id_, _type_):
+        def list_resources(t_ctx, q_ctx, pod_, ele, _type_):
             client = self._get_client(pod_['pod_name'])
             return client.list_resources(_type_, t_ctx, [{'key': 'name',
                                                           'comparator': 'eq',
-                                                          'value': _id_}])
+                                                          'value': ele['id']}])
 
         return t_lock.get_or_create_route(context, None,
-                                          self.project_id, pod, _id, _type,
-                                          list_resources)
+                                          self.project_id, pod, {'id': _id},
+                                          _type, list_resources)
 
-    def _get_create_network_body(self, network):
-        body = {
-            'network': {
-                'tenant_id': self.project_id,
-                'name': network['id'],
-                'admin_state_up': True
-            }
-        }
-        return body
+    def _handle_router(self, context, pod, net):
+        top_client = self._get_client()
 
-    def _get_create_subnet_body(self, subnet, bottom_net_id):
-        body = {
-            'subnet': {
-                'network_id': bottom_net_id,
-                'name': subnet['id'],
-                'ip_version': subnet['ip_version'],
-                'cidr': subnet['cidr'],
-                'gateway_ip': subnet['gateway_ip'],
-                'allocation_pools': subnet['allocation_pools'],
-                'enable_dhcp': subnet['enable_dhcp'],
-                'tenant_id': self.project_id
-            }
-        }
-        return body
-
-    def _get_create_port_body(self, port, subnet_map, bottom_net_id,
-                              security_group_ids=None):
-        bottom_fixed_ips = []
-        for ip in port['fixed_ips']:
-            bottom_ip = {'subnet_id': subnet_map[ip['subnet_id']],
-                         'ip_address': ip['ip_address']}
-            bottom_fixed_ips.append(bottom_ip)
-        body = {
-            'port': {
-                'tenant_id': self.project_id,
-                'admin_state_up': True,
-                'name': port['id'],
-                'network_id': bottom_net_id,
-                'mac_address': port['mac_address'],
-                'fixed_ips': bottom_fixed_ips
-            }
-        }
-        if security_group_ids:
-            body['port']['security_groups'] = security_group_ids
-        return body
-
-    def _get_create_dhcp_port_body(self, port, bottom_subnet_id,
-                                   bottom_net_id):
-        body = {
-            'port': {
-                'tenant_id': self.project_id,
-                'admin_state_up': True,
-                'name': port['id'],
-                'network_id': bottom_net_id,
-                'fixed_ips': [
-                    {'subnet_id': bottom_subnet_id,
-                     'ip_address': port['fixed_ips'][0]['ip_address']}
-                ],
-                'mac_address': port['mac_address'],
-                'binding:profile': {},
-                'device_id': 'reserved_dhcp_port',
-                'device_owner': 'network:dhcp',
-            }
-        }
-        return body
-
-    def _prepare_neutron_element(self, context, pod, ele, _type, body):
-        def list_resources(t_ctx, q_ctx, pod_, _id_, _type_):
-            client = self._get_client(pod_['pod_name'])
-            return client.list_resources(_type_, t_ctx, [{'key': 'name',
-                                                          'comparator': 'eq',
-                                                          'value': _id_}])
-
-        def create_resources(t_ctx, q_ctx, pod_, body_, _type_):
-            client = self._get_client(pod_['pod_name'])
-            return client.create_resources(_type_, t_ctx, body_)
-
-        return t_lock.get_or_create_element(
-            context, None,  # we don't need neutron context, so pass None
-            self.project_id, pod, ele, _type, body,
-            list_resources, create_resources)
+        interfaces = top_client.list_ports(
+            context, filters=[{'key': 'network_id',
+                               'comparator': 'eq',
+                               'value': net['id']},
+                              {'key': 'device_owner',
+                               'comparator': 'eq',
+                               'value': 'network:router_interface'}])
+        interfaces = [inf for inf in interfaces if inf['device_id']]
+        if not interfaces:
+            return
+        # TODO(zhiyuan) change xjob invoking from "cast" to "call" to guarantee
+        # the job can be successfully registered
+        self.xjob_handler.setup_bottom_router(
+            context, net['id'], interfaces[0]['device_id'], pod['pod_id'])
 
     def _handle_network(self, context, pod, net, subnets, port=None,
                         top_sg_ids=None, bottom_sg_ids=None):
-        # network
-        net_body = self._get_create_network_body(net)
-        _, bottom_net_id = self._prepare_neutron_element(context, pod, net,
-                                                         'network', net_body)
+        (bottom_net_id,
+         subnet_map) = self.helper.prepare_bottom_network_subnets(
+            context, None, self.project_id, pod, net, subnets)
 
-        # subnet
-        subnet_map = {}
-        for subnet in subnets:
-            subnet_body = self._get_create_subnet_body(subnet, bottom_net_id)
-            _, bottom_subnet_id = self._prepare_neutron_element(
-                context, pod, subnet, 'subnet', subnet_body)
-            subnet_map[subnet['id']] = bottom_subnet_id
         top_client = self._get_client()
         top_port_body = {'port': {'network_id': net['id'],
                                   'admin_state_up': True}}
         if top_sg_ids:
             top_port_body['port']['security_groups'] = top_sg_ids
 
-        # dhcp port
-        client = self._get_client(pod['pod_name'])
-        t_dhcp_port_filters = [
-            {'key': 'device_owner', 'comparator': 'eq',
-             'value': 'network:dhcp'},
-            {'key': 'network_id', 'comparator': 'eq',
-             'value': net['id']},
-        ]
-        b_dhcp_port_filters = [
-            {'key': 'device_owner', 'comparator': 'eq',
-             'value': 'network:dhcp'},
-            {'key': 'network_id', 'comparator': 'eq',
-             'value': bottom_net_id},
-        ]
-        top_dhcp_port_body = {
-            'port': {
-                'tenant_id': self.project_id,
-                'admin_state_up': True,
-                'name': 'dhcp_port',
-                'network_id': net['id'],
-                'binding:profile': {},
-                'device_id': 'reserved_dhcp_port',
-                'device_owner': 'network:dhcp',
-            }
-        }
-        t_dhcp_ports = top_client.list_ports(context, t_dhcp_port_filters)
-        t_subnet_dhcp_map = {}
-        for dhcp_port in t_dhcp_ports:
-            subnet_id = dhcp_port['fixed_ips'][0]['subnet_id']
-            t_subnet_dhcp_map[subnet_id] = dhcp_port
-        for t_subnet_id, b_subnet_id in subnet_map.iteritems():
-            if t_subnet_id in t_subnet_dhcp_map:
-                t_dhcp_port = t_subnet_dhcp_map[t_subnet_id]
-            else:
-                t_dhcp_port = top_client.create_ports(context,
-                                                      top_dhcp_port_body)
-            mappings = db_api.get_bottom_mappings_by_top_id(
-                context, t_dhcp_port['id'], constants.RT_PORT)
-            pod_list = [mapping[0]['pod_id'] for mapping in mappings]
-            if pod['pod_id'] in pod_list:
-                # mapping exists, skip this subnet
-                continue
-
-            dhcp_port_body = self._get_create_dhcp_port_body(
-                t_dhcp_port, b_subnet_id, bottom_net_id)
-            t_dhcp_ip = t_dhcp_port['fixed_ips'][0]['ip_address']
-
-            b_dhcp_port = None
-            try:
-                b_dhcp_port = client.create_ports(context, dhcp_port_body)
-            except Exception:
-                # examine if we conflicted with a dhcp port which was
-                # automatically created by bottom pod
-                b_dhcp_ports = client.list_ports(context,
-                                                 b_dhcp_port_filters)
-                dhcp_port_match = False
-                for dhcp_port in b_dhcp_ports:
-                    subnet_id = dhcp_port['fixed_ips'][0]['subnet_id']
-                    ip = dhcp_port['fixed_ips'][0]['ip_address']
-                    if b_subnet_id == subnet_id and t_dhcp_ip == ip:
-                        with context.session.begin():
-                            core.create_resource(
-                                context, models.ResourceRouting,
-                                {'top_id': t_dhcp_port['id'],
-                                 'bottom_id': dhcp_port['id'],
-                                 'pod_id': pod['pod_id'],
-                                 'project_id': self.project_id,
-                                 'resource_type': constants.RT_PORT})
-                        dhcp_port_match = True
-                        break
-                if not dhcp_port_match:
-                    # so we didn't conflict with a dhcp port, raise exception
-                    raise
-
-            if b_dhcp_port:
-                with context.session.begin():
-                    core.create_resource(context, models.ResourceRouting,
-                                         {'top_id': t_dhcp_port['id'],
-                                          'bottom_id': b_dhcp_port['id'],
-                                          'pod_id': pod['pod_id'],
-                                          'project_id': self.project_id,
-                                          'resource_type': constants.RT_PORT})
-                # there is still one thing to do, there may be other dhcp ports
-                # created by bottom pod, we need to delete them
-                b_dhcp_ports = client.list_ports(context,
-                                                 b_dhcp_port_filters)
-                remove_port_list = []
-                for dhcp_port in b_dhcp_ports:
-                    subnet_id = dhcp_port['fixed_ips'][0]['subnet_id']
-                    ip = dhcp_port['fixed_ips'][0]['ip_address']
-                    if b_subnet_id == subnet_id and t_dhcp_ip != ip:
-                        remove_port_list.append(dhcp_port['id'])
-                for dhcp_port_id in remove_port_list:
-                    # NOTE(zhiyuan) dhcp agent will receive this port-delete
-                    # notification and re-configure dhcp so our newly created
-                    # dhcp port can be used
-                    client.delete_ports(context, dhcp_port_id)
-
         # port
         if not port:
             port = top_client.create_ports(context, top_port_body)
-            port_body = self._get_create_port_body(
-                port, subnet_map, bottom_net_id, bottom_sg_ids)
+            port_body = self.helper.get_create_port_body(
+                self.project_id, port, subnet_map, bottom_net_id,
+                bottom_sg_ids)
         else:
-            port_body = self._get_create_port_body(port, subnet_map,
-                                                   bottom_net_id)
-        _, bottom_port_id = self._prepare_neutron_element(context, pod, port,
-                                                          'port', port_body)
+            port_body = self.helper.get_create_port_body(
+                self.project_id, port, subnet_map, bottom_net_id)
+        _, bottom_port_id = self.helper.prepare_bottom_element(
+            context, self.project_id, pod, port, constants.RT_PORT, port_body)
+
+        self._handle_router(context, pod, net)
+
         return port['id'], bottom_port_id
 
     def _handle_port(self, context, pod, port):
@@ -479,8 +389,8 @@ class ServerController(rest.RestController):
                 'security_group': {
                     'name': t_sg['id'],
                     'description': t_sg['description']}}
-            is_new, b_sg_id = self._prepare_neutron_element(
-                context, pod, t_sg, constants.RT_SG, sg_body)
+            is_new, b_sg_id = self.helper.prepare_bottom_element(
+                context, self.project_id, pod, t_sg, constants.RT_SG, sg_body)
             t_sg_ids.append(t_sg['id'])
             is_news.append(is_new)
             b_sg_ids.append(b_sg_id)
@@ -636,15 +546,36 @@ class ServerController(rest.RestController):
                     del addresses[remove_index]
 
     @staticmethod
-    def _check_network_server_the_same_az(network, server_az):
+    def _remove_stale_mapping(context, server_id):
+        filters = [{'key': 'top_id', 'comparator': 'eq', 'value': server_id},
+                   {'key': 'resource_type',
+                    'comparator': 'eq',
+                    'value': constants.RT_SERVER}]
+        with context.session.begin():
+            core.delete_resources(context,
+                                  models.ResourceRouting,
+                                  filters)
+
+    @staticmethod
+    def _check_network_server_az_match(context, network, server_az):
         az_hints = 'availability_zone_hints'
+        network_type = 'provider:network_type'
+
+        # for local type network, we make sure it's created in only one az
+
+        # NOTE(zhiyuan) race condition exists when creating vms in the same
+        # local type network but different azs at the same time
+        if network.get(network_type) == constants.NT_LOCAL:
+            mappings = db_api.get_bottom_mappings_by_top_id(
+                context, network['id'], constants.RT_NETWORK)
+            if mappings:
+                pod, _ = mappings[0]
+                if pod['az_name'] != server_az:
+                    return False
         # if neutron az not assigned, server az is used
         if not network.get(az_hints):
             return True
-        # temporally not support cross-pod network
-        if len(network[az_hints]) > 1:
-            return False
-        if network[az_hints][0] == server_az:
+        if server_az in network[az_hints]:
             return True
         else:
             return False
@@ -660,7 +591,7 @@ class ServerController(rest.RestController):
             msg = str(e)
             LOG.exception(_LE('Quota exceeded %(msg)s'),
                           {'msg': msg})
-            pecan.abort(400, _('Quota exceeded %s') % msg)
+            return utils.format_nova_error(400, _('Quota exceeded %s') % msg)
 
     def _check_injected_file_quota(self, context, injected_files):
         """Enforce quota limits on injected files.
@@ -707,15 +638,16 @@ class ServerController(rest.RestController):
         except t_exceptions.InvalidMetadata as e1:
             LOG.exception(_LE('Invalid metadata %(exception)s'),
                           {'exception': str(e1)})
-            pecan.abort(400, _('Invalid metadata'))
+            return utils.format_nova_error(400, _('Invalid metadata'))
         except t_exceptions.InvalidMetadataSize as e2:
             LOG.exception(_LE('Invalid metadata size %(exception)s'),
                           {'exception': str(e2)})
-            pecan.abort(400, _('Invalid metadata size'))
+            return utils.format_nova_error(400, _('Invalid metadata size'))
         except t_exceptions.MetadataLimitExceeded as e3:
             LOG.exception(_LE('Quota exceeded %(exception)s'),
                           {'exception': str(e3)})
-            pecan.abort(400, _('Quota exceeded in metadata'))
+            return utils.format_nova_error(400,
+                                           _('Quota exceeded in metadata'))
 
     def _check_metadata_properties_quota(self, context, metadata=None):
         """Enforce quota limits on metadata properties."""
